from sqlalchemy import Engine, select
from sqlalchemy.orm import sessionmaker

from .models.config import *
from .models.scores import *
from .models.songs import *
from .singleton import Singleton


class Database(metaclass=Singleton):
    def __init__(self, engine: Engine):
        self.engine = engine

    @property
    def engine(self):
        return self.__engine

    @engine.setter
    def engine(self, value: Engine):
        if not isinstance(value, Engine):
            raise ValueError("Database.engine only accepts sqlalchemy.Engine")
        self.__engine = value
        self.__sessionmaker = sessionmaker(self.__engine)

    @property
    def sessionmaker(self):
        return self.__sessionmaker

    def init(self, checkfirst: bool = True):
        # create tables & views
        if checkfirst:
            # > https://github.com/kvesteri/sqlalchemy-utils/issues/396
            # > view.create_view() causes DuplicateTableError on Base.metadata.create_all(checkfirst=True)
            # so if `checkfirst` is True, drop these views before creating
            ScoresViewBase.metadata.drop_all(self.engine)

        SongsBase.metadata.create_all(self.engine, checkfirst=checkfirst)
        ScoresBase.metadata.create_all(self.engine, checkfirst=checkfirst)
        ScoresViewBase.metadata.create_all(self.engine)
        ConfigBase.metadata.create_all(self.engine, checkfirst=checkfirst)

        # insert version property
        with self.sessionmaker() as session:
            stmt = select(Property.value).where(Property.key == "version")
            result = session.execute(stmt).fetchone()
            if not checkfirst or not result:
                session.add(Property(key="version", value="2"))
                session.commit()



from typing import Optional

from sqlalchemy import TEXT, ForeignKey
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from .common import ReprHelper

__all__ = [
    "SongsBase",
    "Pack",
    "PackLocalized",
    "Song",
    "SongLocalized",
    "Chart",
    "ChartLocalized",
    "ChartInfo",
]


class SongsBase(DeclarativeBase, ReprHelper):
    pass


class Pack(SongsBase):
    __tablename__ = "pack"

    id: Mapped[str] = mapped_column(TEXT(), primary_key=True)
    name: Mapped[str] = mapped_column(TEXT())
    description: Mapped[Optional[str]] = mapped_column(TEXT())


class PackLocalized(SongsBase):
    __tablename__ = "pack_localized"

    id: Mapped[str] = mapped_column(ForeignKey("pack.id"), primary_key=True)
    name_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    name_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    name_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    name_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())
    description_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    description_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    description_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    description_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())


class Song(SongsBase):
    __tablename__ = "song"

    idx: Mapped[int]
    id: Mapped[str] = mapped_column(TEXT(), primary_key=True)
    title: Mapped[str] = mapped_column(TEXT())
    artist: Mapped[str] = mapped_column(TEXT())
    set: Mapped[str] = mapped_column(TEXT())
    bpm: Mapped[Optional[str]] = mapped_column(TEXT())
    bpm_base: Mapped[Optional[float]]
    audio_preview: Mapped[Optional[int]]
    audio_preview_end: Mapped[Optional[int]]
    side: Mapped[Optional[int]]
    version: Mapped[Optional[str]] = mapped_column(TEXT())
    date: Mapped[Optional[int]]
    bg: Mapped[Optional[str]] = mapped_column(TEXT())
    bg_inverse: Mapped[Optional[str]] = mapped_column(TEXT())
    bg_day: Mapped[Optional[str]] = mapped_column(TEXT())
    bg_night: Mapped[Optional[str]] = mapped_column(TEXT())
    source: Mapped[Optional[str]] = mapped_column(TEXT())
    source_copyright: Mapped[Optional[str]] = mapped_column(TEXT())


class SongLocalized(SongsBase):
    __tablename__ = "song_localized"

    id: Mapped[str] = mapped_column(ForeignKey("song.id"), primary_key=True)
    title_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    title_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    title_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    title_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())
    search_title_ja: Mapped[Optional[str]] = mapped_column(TEXT(), comment="json")
    search_title_ko: Mapped[Optional[str]] = mapped_column(TEXT(), comment="json")
    search_title_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT(), comment="json")
    search_title_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT(), comment="json")
    search_artist_ja: Mapped[Optional[str]] = mapped_column(TEXT(), comment="json")
    search_artist_ko: Mapped[Optional[str]] = mapped_column(TEXT(), comment="json")
    search_artist_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT(), comment="json")
    search_artist_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT(), comment="json")
    source_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    source_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    source_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    source_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())


class Chart(SongsBase):
    __tablename__ = "chart"

    song_id: Mapped[str] = mapped_column(TEXT(), primary_key=True)
    rating_class: Mapped[int] = mapped_column(primary_key=True)
    rating: Mapped[int]
    rating_plus: Mapped[bool]
    chart_designer: Mapped[Optional[str]] = mapped_column(TEXT())
    jacket_desginer: Mapped[Optional[str]] = mapped_column(TEXT())
    audio_override: Mapped[bool]
    jacket_override: Mapped[bool]
    jacket_night: Mapped[Optional[str]] = mapped_column(TEXT())
    title: Mapped[Optional[str]] = mapped_column(TEXT())
    artist: Mapped[Optional[str]] = mapped_column(TEXT())
    bg: Mapped[Optional[str]] = mapped_column(TEXT())
    bg_inverse: Mapped[Optional[str]] = mapped_column(TEXT())
    bpm: Mapped[Optional[str]] = mapped_column(TEXT())
    bpm_base: Mapped[Optional[float]]
    version: Mapped[Optional[str]] = mapped_column(TEXT())
    date: Mapped[Optional[int]]


class ChartLocalized(SongsBase):
    __tablename__ = "chart_localized"

    song_id: Mapped[str] = mapped_column(ForeignKey("chart.song_id"), primary_key=True)
    rating_class: Mapped[str] = mapped_column(
        ForeignKey("chart.rating_class"), primary_key=True
    )
    title_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    title_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    title_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    title_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())
    artist_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    artist_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    artist_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    artist_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())


class ChartInfo(SongsBase):
    __tablename__ = "chart_info"

    song_id: Mapped[str] = mapped_column(ForeignKey("chart.song_id"), primary_key=True)
    rating_class: Mapped[str] = mapped_column(
        ForeignKey("chart.rating_class"), primary_key=True
    )
    constant: Mapped[int] = mapped_column(
        comment="real_constant * 10. For example, Crimson Throne [FTR] is 10.4, then store 104 here."
    )
    note: Mapped[Optional[int]]



from decimal import Decimal
from math import floor
from typing import Dict, List

from .models.scores import Calculated


def calculate_score_range(notes: int, pure: int, far: int):
    single_note_score = 10000000 / Decimal(notes)

    actual_score = floor(
        single_note_score * pure + single_note_score * Decimal(0.5) * far
    )
    return (actual_score, actual_score + pure)


def calculate_potential(_constant: float, score: int) -> Decimal:
    constant = Decimal(_constant)
    if score >= 10000000:
        return constant + 2
    elif score >= 9800000:
        return constant + 1 + (Decimal(score - 9800000) / 200000)
    else:
        return max(Decimal(0), constant + (Decimal(score - 9500000) / 300000))


def calculate_shiny_pure(notes: int, score: int, pure: int, far: int) -> int:
    single_note_score = 10000000 / Decimal(notes)
    actual_score = single_note_score * pure + single_note_score * Decimal(0.5) * far
    return score - floor(actual_score)


def get_b30_calculated_list(calculated_list: List[Calculated]) -> List[Calculated]:
    best_scores: Dict[str, Calculated] = {}
    for calculated in calculated_list:
        key = f"{calculated.song_id}_{calculated.rating_class}"
        stored = best_scores.get(key)
        if stored and stored.score < calculated.score or not stored:
            best_scores[key] = calculated
    ret_list = list(best_scores.values())
    ret_list = sorted(ret_list, key=lambda c: c.potential, reverse=True)[:30]
    return ret_list


def calculate_b30(calculated_list: List[Calculated]) -> Decimal:
    ptt_list = [Decimal(c.potential) for c in get_b30_calculated_list(calculated_list)]
    sum_ptt_list = sum(ptt_list)
    return (sum_ptt_list / len(ptt_list)) if sum_ptt_list else Decimal("0.0")



import json
from os import PathLike
from typing import Any, List, Optional, Union

from sqlalchemy.orm import DeclarativeBase, Session


def to_db_value(val: Any) -> Any:
    if not val:
        return None
    elif isinstance(val, list):
        return json.dumps(val, ensure_ascii=False)
    else:
        return val


def is_localized(item: dict, key: str, append_localized: bool = True):
    item_key = f"{key}_localized" if append_localized else key
    subitem: Optional[dict] = item.get(item_key)
    return subitem and (
        subitem.get("ja")
        or subitem.get("ko")
        or subitem.get("zh-Hant")
        or subitem.get("zh-Hans")
    )


def set_model_localized_attrs(
    model: DeclarativeBase, item: dict, model_key: str, item_key: Optional[str] = None
):
    if item_key is None:
        item_key = f"{model_key}_localized"
    subitem: dict = item.get(item_key, {})
    if not subitem:
        return
    setattr(model, f"{model_key}_ja", to_db_value(subitem.get("ja")))
    setattr(model, f"{model_key}_ko", to_db_value(subitem.get("ko")))
    setattr(model, f"{model_key}_zh_hans", to_db_value(subitem.get("zh-Hans")))
    setattr(model, f"{model_key}_zh_hant", to_db_value(subitem.get("zh-Hant")))


class ArcaeaParser:
    def __init__(self, filepath: Union[str, bytes, PathLike]):
        self.filepath = filepath

    def parse(self) -> List[DeclarativeBase]:
        ...

    def write_database(self, session: Session):
        results = self.parse()
        for result in results:
            session.merge(result)



import logging
import sqlite3
from typing import List

from sqlalchemy import select
from sqlalchemy.orm import Session

from ...models.scores import Score
from .common import ArcaeaParser

logger = logging.getLogger(__name__)


class St3ScoreParser(ArcaeaParser):
    CLEAR_TYPES_MAP = {0: -1, 1: 1, 2: 1, 3: 1, 4: 1, 5: 1}

    def __init__(self, filepath):
        super().__init__(filepath)

    def parse(self) -> List[Score]:
        items = []
        with sqlite3.connect(self.filepath) as st3_conn:
            cursor = st3_conn.cursor()
            db_scores = cursor.execute(
                "SELECT songId, songDifficulty, score, perfectCount, nearCount, missCount, date FROM scores"
            ).fetchall()
            for song_id, rating_class, score, pure, far, lost, date in db_scores:
                db_clear_type = cursor.execute(
                    "SELECT clearType FROM cleartypes WHERE songId = ? AND songDifficulty = ?",
                    (song_id, rating_class),
                ).fetchone()[0]
                r10_clear_type = self.CLEAR_TYPES_MAP[db_clear_type]

                date_str = str(date)
                date = None if len(date_str) < 7 else int(date_str.ljust(10, "0"))

                items.append(
                    Score(
                        song_id=song_id,
                        rating_class=rating_class,
                        score=score,
                        pure=pure,
                        far=far,
                        lost=lost,
                        date=date,
                        r10_clear_type=r10_clear_type,
                    )
                )

        return items

    def write_database(self, session: Session, *, skip_duplicate=True):
        parsed_scores = self.parse()
        for parsed_score in parsed_scores:
            query_score = session.scalar(
                select(Score).where(
                    (Score.song_id == parsed_score.song_id)
                    & (Score.rating_class == parsed_score.rating_class)
                    & (Score.score == parsed_score.score)
                )
            )

            if query_score and skip_duplicate:
                logger.info(
                    f"{repr(parsed_score)} skipped because "
                    f"potential duplicate item {repr(query_score)} found."
                )
                continue
            session.add(parsed_score)
